from flowjax.bijections.abc import ParameterisedBijection
import equinox as eqx
import jax.numpy as jnp
from jax import random
from flowjax.bijections.abc import Bijection


class Coupling(Bijection, eqx.Module):
    d: int
    D: int
    bijection: ParameterisedBijection
    conditioner: eqx.nn.MLP
    condition_dim: int

    def __init__(
        self,
        key: random.PRNGKey,
        bijection: ParameterisedBijection,
        d: int,
        D: int,
        condition_dim: int,
        nn_width: int,
        nn_depth: int,
    ):
        """Coupling layer implementation.

        Args:
            key (random.PRNGKey): Key
            bijection (ParameterisedBijection): Bijection to be parameterised by the conditioner neural netork.
            d (int): Number of untransformed conditioning variables.
            D (int): Total dimension.
            nn_width (int): Number of nodes in hidden layers.
            nn_depth (int): Number of hidden layers.
        """
        self.d = d
        self.D = D
        self.bijection = bijection
        output_size = self.bijection.num_params(D - d)

        self.conditioner = eqx.nn.MLP(
            in_size=d + condition_dim,
            out_size=output_size,
            width_size=nn_width,
            depth=nn_depth,
            key=key,
        )

        self.condition_dim = condition_dim

    def transform(self, x: jnp.ndarray, condition=None):
        condition = jnp.array([]) if condition is None else condition
        x_cond, x_trans = x[: self.d], x[self.d :]
        cond = jnp.concatenate((x_cond, condition))
        bijection_params = self.conditioner(cond)
        bijection_args = self.bijection.get_args(bijection_params)
        y_trans = self.bijection.transform(x_trans, *bijection_args)
        y = jnp.concatenate((x_cond, y_trans))
        return y

    def transform_and_log_abs_det_jacobian(self, x: jnp.ndarray, condition=None):
        condition = jnp.array([]) if condition is None else condition
        x_cond, x_trans = x[: self.d], x[self.d :]
        cond = jnp.concatenate((x_cond, condition))

        bijection_params = self.conditioner(cond)
        bijection_args = self.bijection.get_args(bijection_params)
        y_trans, log_abs_det = self.bijection.transform_and_log_abs_det_jacobian(
            x_trans, *bijection_args
        )
        y = jnp.concatenate([x_cond, y_trans])
        return y, log_abs_det

    def inverse(self, y: jnp.ndarray, condition=None):
        condition = jnp.array([]) if condition is None else condition
        x_cond, y_trans = y[: self.d], y[self.d :]
        cond = jnp.concatenate((x_cond, condition))
        bijection_params = self.conditioner(cond)
        bijection_args = self.bijection.get_args(bijection_params)
        x_trans = self.bijection.inverse(y_trans, *bijection_args)
        x = jnp.concatenate((x_cond, x_trans))
        return x



from typing import Callable, Optional
import jax.numpy as jnp
from numpy import broadcast_shapes
from flowjax.bijections.abc import Bijection
from jax.scipy.stats import norm
from jax import random
import equinox as eqx
import jax
from flowjax.bijections.coupling import Coupling
from flowjax.bijections.rational_quadratic_spline import RationalQuadraticSpline
from flowjax.bijections.affine import Affine
from flowjax.bijections.utils import Chain
from flowjax.bijections.bnaf import BlockAutoregressiveNetwork
from flowjax.bijections.utils import intertwine_permute


class Flow(eqx.Module):
    bijection: Bijection
    target_dim: int
    base_log_prob: Callable
    base_sample: Callable

    def __init__(
        self,
        bijection: Bijection,
        target_dim: int,
        base_log_prob: Optional[Callable] = None,
        base_sample: Optional[Callable] = None,
    ):
        """Form a distribution like object using a base distribution and a
        bijection. Operations are generally assumed to be batched along
        dimension 0.

        Args:
            bijection (Bijection): Bijection mapping from target distribution to
                the base distribution.
            target_dim: (int): Dimension of the target distribution.
            base_log_prob (Optional[Callable], optional): log probability in the base
                distribution. Defaults to standard normal.
            base_sample (Optional[Callable], optional): sample function with signature
                (key : PRNGKey, n : int). Defaults to standard normal.
        """
        self.bijection = bijection
        self.target_dim = target_dim

        if base_log_prob:
            self.base_log_prob = base_log_prob
            self.base_sample = base_sample
        else:
            self.base_log_prob = lambda x: norm.logpdf(x).sum(axis=1)
            self.base_sample = lambda key, n: random.normal(key, (n, target_dim))

    @eqx.filter_jit
    def log_prob(self, x: jnp.ndarray, condition: Optional[jnp.ndarray] = None):
        "Evaluate the log probability of the target distribution."
        x = jnp.atleast_2d(x)

        if condition is not None:
            x, condition = self._broadcast(x, condition)

        z, log_abs_det = jax.vmap(self.bijection.transform_and_log_abs_det_jacobian)(
            x, condition
        )
        p_z = self.base_log_prob(z)
        return p_z + log_abs_det

    @staticmethod
    def _broadcast(x: jnp.ndarray, condition: jnp.ndarray):
        "Broadcast arrays, excluding last axis."
        s = broadcast_shapes(x.shape[:-1], condition.shape[:-1])
        x = jnp.broadcast_to(x, s + (x.shape[-1],))
        condition = jnp.broadcast_to(condition, s + (condition.shape[-1],))
        return x, condition

    @eqx.filter_jit
    def sample(
        self,
        key: random.PRNGKey,
        condition: Optional[jnp.ndarray] = None,
        n: Optional[int] = None,
    ):
        """Sample from the (conditional or unconditional) flow. For repeated sampling using
        a particular instance of the conditioning variable, use a vector condition and n to
        specify the number of samples. To sample once for may different conditioning variables
        provide a matrix of conditioning variables (n is inferred from axis 0).

        Args:
            key (random.PRNGKey): Random key.
            condition (jnp.ndarray, optional): Conditioning variables. Defaults to None.
            n (Optional[int], optional): Number of samples. Defaults to None.

        Returns:
            jnp.ndarray: Samples from the target distribution.
        """

        if condition is not None:
            if condition.ndim == 1:
                assert n is not None, "n must be provided with a vector condition."
                condition = jnp.broadcast_to(condition, (n, condition.shape[0]))
            else:
                assert n is None, "n should not be provided if a matrix of conditioning variables is used."
                n = condition.shape[0]

        z = self.base_sample(key, n)
        x = jax.vmap(self.bijection.inverse)(z, condition)
        return x


class NeuralSplineFlow(Flow):
    def __init__(
        self,
        key: random.PRNGKey,
        target_dim: int,
        condition_dim: int = 0,
        K: int = 10,
        B: int = 5,
        num_layers: int = 5,
        nn_width: int = 40,
        nn_depth: int = 2,
        permute_strategy: Optional[str] = None,
        base_log_prob: Optional[Callable] = None,
        base_sample: Optional[Callable] = None,
    ):
        """Convenience constructor for Neural spline flow (Durkan et al. 2019;
        https://arxiv.org/abs/1906.04032). Note that the transformation is on
        the interval [-B, B].

        Args:
            key (random.PRNGKey): Random key.
            target_dim (int): Dimension of the target distribution.
            condition_dim (int, optional): Dimension of extra conditioning variables. Defaults to 0.
            K (int, optional): Number of (inner) spline segments. Defaults to 10.
            B (int, optional): Interval to transform [-B, B]. Defaults to 5.
            num_layers (int, optional): Number of coupling layers. Defaults to 5.
            nn_width (int, optional): Conditioner network width. Defaults to 40.
            nn_depth (int, optional): Conditioner network depth. Defaults to 2.
            permute_strategy (Optional[str], optional): How to permute between layers. Either "flip" or "random". Defaults to "flip" if target_dim <=2, otherwise "random".
            base_log_prob (Optional[Callable], optional): Log probability in base distribution. Defaults to standard normal.
            base_sample (Optional[Callable], optional): Sample function in base distribution. Defaults to standard normal.
        """
        d = target_dim // 2
        if permute_strategy is None:
            permute_strategy = "flip" if target_dim <= 2 else "random"

        permute_key, *layer_keys = random.split(key, num_layers + 1)
        layers = [
            Coupling(
                key=key,
                bijection=RationalQuadraticSpline(K=K, B=B),
                d=d,
                D=target_dim,
                nn_width=nn_width,
                nn_depth=nn_depth,
                condition_dim=condition_dim,
            )
            for key in layer_keys
        ]

        layers = intertwine_permute(layers, permute_strategy, permute_key, target_dim)
        bijection = Chain(layers)

        super().__init__(bijection, target_dim, base_log_prob, base_sample)


class RealNVPFlow(Flow):
    def __init__(
        self,
        key: random.PRNGKey,
        target_dim: int,
        condition_dim: int = 0,
        num_layers: int = 5,
        nn_width: int = 40,
        nn_depth: int = 2,
        permute_strategy: Optional[str] = None,
        base_log_prob: Optional[Callable] = None,
        base_sample: Optional[Callable] = None,
    ):
        """Convenience constructor for a RealNVP style flow (Dinh et al, 2017;
        https://arxiv.org/abs/1605.08803). Note this implementation differs slightly
        from the original, e.g. it does not use batch normaliziation.

        Args:
            key (random.PRNGKey): Random key.
            target_dim (int): Dimension of the target distribution.
            condition_dim (int, optional): Dimension of extra conditioning variables. Defaults to 0.
            num_layers (int, optional): Number of coupling layers. Defaults to 5.
            nn_width (int, optional): Conditioner network width. Defaults to 40.
            nn_depth (int, optional): Conditioner network depth. Defaults to 2.
            permute_strategy (Optional[str], optional): How to permute between layers. Either "flip" or "random". Defaults to "flip" if target_dim <=2, otherwise "random".
            base_log_prob (Optional[Callable], optional): Log probability in base distribution. Defaults to standard normal.
            base_sample (Optional[Callable], optional): Sample function in base distribution. Defaults to standard normal.
        """
        d = target_dim // 2
        if permute_strategy is None:
            permute_strategy = "flip" if target_dim <= 2 else "random"

        permute_key, *layer_keys = random.split(key, num_layers + 1)
        layers = [
            Coupling(
                key=key,
                bijection=Affine(),
                d=d,
                D=target_dim,
                nn_width=nn_width,
                nn_depth=nn_depth,
                condition_dim=condition_dim,
            )
            for key in layer_keys
        ]

        layers = intertwine_permute(layers, permute_strategy, permute_key, target_dim)
        bijection = Chain(layers)
        super().__init__(bijection, target_dim, base_log_prob, base_sample)


class BlockNeuralAutoregressiveFlow(Flow):
    def __init__(
        self,
        key: random.PRNGKey,
        target_dim: int,
        condition_dim: int = 0,
        nn_layers: int = 3,
        block_size: tuple = (8, 8),
        flow_layers: int = 1,
        permute_strategy: Optional[str] = None,
        base_log_prob: Optional[Callable] = None,
        base_sample: Optional[Callable] = None,
    ):
        """Convenience constructor for a block neural autoregressive flow
        (https://arxiv.org/abs/1904.04676).

        Args:
            key (random.PRNGKey): Random key.
            target_dim (int): Dimension of the target distribution.
            nn_layers (int, optional): Number of layers within autoregressive neural networks. Defaults to 3.
            block_size (tuple, optional): Block size in lower triangular blocks of autoregressive neural network. Defaults to (8, 8).
            flow_layers (int, optional): Number of flow layers (1 layer = autoregressive neural network + TanH activation) . Defaults to 1.
            permute_strategy (Optional[str], optional): How to permute between layers. Either "flip" or "random". Defaults to "flip" if target_dim <=2, otherwise "random".
            base_log_prob (Optional[Callable], optional): Base distribution log probability function. Defaults to standard normal.
            base_sample (Optional[Callable], optional): Base distribution sample function. Defaults to standard normal.
        """
        assert nn_layers >= 2

        if permute_strategy is None:
            permute_strategy = "flip" if target_dim <= 2 else "random"

        permute_key, *layer_keys = random.split(key, flow_layers + 1)

        bijections = [
            BlockAutoregressiveNetwork(
                key, dim=target_dim, condition_dim=condition_dim,
                n_layers=nn_layers, block_size=block_size
            )
            for key in layer_keys
        ]

        bijections = intertwine_permute(
            bijections, permute_strategy, permute_key, target_dim
        )
        bijection = Chain(bijections)
        super().__init__(bijection, target_dim, base_log_prob, base_sample)



from flowjax.bijections.abc import ParameterisedBijection
import jax.numpy as jnp
import jax
from functools import partial


class RationalQuadraticSpline(ParameterisedBijection):

    def __init__(self, K, B, min_bin_width=1e-3, min_bin_height=1e-3, min_derivative=1e-3):
        self.K = K
        self.B = B
        self.min_bin_width = min_bin_width
        self.min_bin_height = min_bin_height
        self.min_derivative = min_derivative
        pos_pad = jnp.zeros(self.K + 4)
        pad_idxs = jnp.array([0, 1, -2, -1])
        pad_vals = jnp.array(
            [-B * 1e4, -B, B, B * 1e4]
        )  # Avoids jax control flow for identity tails
        """
        RationalQuadraticSpline following Durkin et al. (2019),
        https://arxiv.org/abs/1906.04032. Each row of parameter matrices (x_pos,
        y_pos, derivatives) corresponds to a column (axis=1) in x. Ouside the interval
        [-B, B], the identity transform is used.

        Args:
            K (int): Number of inner knots B: (int):
            B: (int): Interval to transform [-B, B]
        """
        pos_pad = pos_pad.at[pad_idxs].set(pad_vals)
        self._pos_pad = pos_pad # End knots and beyond

    @property
    def pos_pad(self):
        return jax.lax.stop_gradient(self._pos_pad)

    @partial(jax.vmap, in_axes=[None, 0, 0, 0, 0])
    def transform(self, x, x_pos, y_pos, derivatives):
        k = self._get_bin(x, x_pos)
        yk, yk1, xk, xk1 = y_pos[k], y_pos[k + 1], x_pos[k], x_pos[k + 1]
        dk, dk1 = derivatives[k], derivatives[k + 1]
        sk = (yk1 - yk) / (xk1 - xk)
        xi = (x - xk) / (xk1 - xk)
        return self._rational_quadratic(sk, xi, dk, dk1, yk, yk1)

    @partial(jax.vmap, in_axes=[None, 0, 0, 0, 0])
    def inverse(self, y, x_pos, y_pos, derivatives):
        k = self._get_bin(y, y_pos)
        xk, xk1, yk = x_pos[k], x_pos[k + 1], y_pos[k]
        sk = (y_pos[k + 1] - yk) / (xk1 - xk)
        y_delta_s_term = (y - yk) * (derivatives[k + 1] + derivatives[k] - 2 * sk)
        a = (y_pos[k + 1] - yk) * (sk - derivatives[k]) + y_delta_s_term
        b = (y_pos[k + 1] - yk) * derivatives[k] - y_delta_s_term
        c = -sk * (y - yk)
        sqrt_term = jnp.sqrt(b ** 2 - 4 * a * c)
        xi = (2 * c) / (-b - sqrt_term)
        x = xi * (xk1 - xk) + xk
        return x

    def transform_and_log_abs_det_jacobian(self, x, x_pos, y_pos, derivatives):
        y, log_det = jax.vmap(self._transform_and_log_abs_det_jacobian)(
            x, x_pos, y_pos, derivatives
        )
        return y, log_det.sum()

    # Methods prepended with _ are defined for scalar x, and are vmapped as appropriate.
    def _transform_and_log_abs_det_jacobian(self, x, x_pos, y_pos, derivatives):
        "Defined for single dimensional x."
        k = self._get_bin(x, x_pos)
        yk, yk1, xk, xk1 = y_pos[k], y_pos[k + 1], x_pos[k], x_pos[k + 1]
        dk, dk1 = derivatives[k], derivatives[k + 1]
        sk = (yk1 - yk) / (xk1 - xk)
        xi = (x - xk) / (xk1 - xk)
        y = self._rational_quadratic(sk, xi, dk, dk1, yk, yk1)
        derivative = self._derivative(sk, xi, dk, dk1)
        return y, jnp.log(derivative)

    def num_params(self, dim: int):
        return (self.K * 3 - 1) * dim

    def get_args(self, params):
        params = params.reshape((-1, self.K * 3 - 1))
        return jax.vmap(self._get_args)(params)

    def _get_args(self, params):
        "Gets the arguments for a single dimension of x."
        widths = jax.nn.softmax(params[: self.K]) * 2 * self.B
        widths = self.min_bin_width + (1 - self.min_bin_width * self.K) * widths

        heights = jax.nn.softmax(params[self.K : self.K * 2]) * 2 * self.B
        heights = self.min_bin_height + (1 - self.min_bin_height * self.K) * heights

        derivatives = jax.nn.softplus(params[self.K * 2 :]) + self.min_derivative
        derivatives = jnp.pad(derivatives, 2, constant_values=1)
        x_pos = jnp.cumsum(widths) - self.B
        x_pos = self.pos_pad.at[2:-2].set(x_pos)
        y_pos = jnp.cumsum(heights) - self.B
        y_pos = self.pos_pad.at[2:-2].set(y_pos)
        return x_pos, y_pos, derivatives

    @staticmethod
    def _get_bin(target, positions):
        "Finds which bin/spline segment the target is in (defined for 1d)."
        cond1 = target <= positions[1:]
        cond2 = target > positions[:-1]
        return jnp.where(cond1 & cond2, size=1)[0][0]

    @staticmethod
    def _rational_quadratic(sk, xi, dk, dk1, yk, yk1):  # eq. 4
        num = (yk1 - yk) * (sk * xi ** 2 + dk * xi * (1 - xi))
        den = sk + (dk1 + dk - 2 * sk) * xi * (1 - xi)
        return yk + num / den

    @staticmethod
    def _derivative(sk, xi, dk, dk1):  # eq. 5
        num = sk ** 2 * (dk1 * xi ** 2 + 2 * sk * xi * (1 - xi) + dk * (1 - xi) ** 2)
        den = (sk + (dk1 + dk - 2 * sk) * xi * (1 - xi)) ** 2
        return num / den



from typing import Callable, Optional
import equinox as eqx
import jax
import jax.numpy as jnp
from flowjax.bijections.abc import Bijection
from jax import random
from jax.nn.initializers import glorot_uniform
from jax import lax


def b_diag_mask(block_shape: tuple, n_blocks: int):
    "Block diagonal mask."
    return jax.scipy.linalg.block_diag(
        *[jnp.ones(block_shape, int) for _ in range(n_blocks)]
    )


def b_tril_mask(block_shape: tuple, n_blocks: int):
    "Upper triangular block mask, excluding diagonal blocks."
    mask = jnp.zeros((block_shape[0] * n_blocks, block_shape[1] * n_blocks))

    for i in range(n_blocks):
        mask = mask.at[
            (i + 1) * block_shape[0] :, i * block_shape[1] : (i + 1) * block_shape[1]
        ].set(1)
    return mask


class BlockAutoregressiveLinear(eqx.Module):
    n_blocks: int
    block_shape: tuple
    condition_dim: int
    W: jnp.ndarray
    bias: jnp.ndarray
    W_log_scale: jnp.ndarray
    in_features: int
    out_features: int
    _b_diag_mask: jnp.ndarray
    _b_diag_mask_idxs: jnp.ndarray
    _b_tril_mask: jnp.ndarray

    def __init__(
        self,
        key: random.PRNGKey,
        n_blocks: int,
        block_shape: tuple,
        condition_dim: int = 0,
        init: Callable = glorot_uniform(),
    ):
        """Block autoregressive neural netork layer (https://arxiv.org/abs/1904.04676).
        Conditioning variables are incorporated by appending columns (one for each
        conditioning variable) to the left of the block diagonal weight matrix.

        Args:
            key (random.PRNGKey): Random key
            n_blocks (int): Number of diagonal blocks (dimension of input layer).
            block_shape (tuple): The shape of the (unconstrained) blocks.
            condition_dim (int): Number of additional conditioning variables. Defaults to 0.
            init (Callable, optional): Default initialisation method for the weight matrix. Defaults to glorot_uniform().
        """
        cond_size = (block_shape[0] * n_blocks, condition_dim)

        self._b_diag_mask = jnp.column_stack(
            (jnp.zeros(cond_size), b_diag_mask(block_shape, n_blocks))
            )

        self._b_tril_mask = jnp.column_stack(
            (jnp.ones(cond_size), b_tril_mask(block_shape, n_blocks))
            )
        self._b_diag_mask_idxs = jnp.where(self._b_diag_mask)

        in_features, out_features = (
            block_shape[1] * n_blocks + condition_dim,
            block_shape[0] * n_blocks,
        )

        *w_key, bias_key, scale_key = random.split(key, n_blocks + 2)

        self.W = init(w_key[0], (out_features, in_features)) * (
            self.b_tril_mask + self.b_diag_mask
        )
        self.bias = (random.uniform(bias_key, (out_features,)) - 0.5) * (
            2 / jnp.sqrt(out_features)
        )

        self.n_blocks = n_blocks
        self.block_shape = block_shape
        self.condition_dim = condition_dim
        self.W_log_scale = jnp.log(random.uniform(scale_key, (out_features, 1)))
        self.in_features = in_features
        self.out_features = out_features

    def get_normalised_weights(self):
        "Carries out weight normalisation."
        W = jnp.exp(self.W) * self.b_diag_mask + self.W * self.b_tril_mask
        W_norms = jnp.linalg.norm(W, axis=-1, keepdims=True)
        return jnp.exp(self.W_log_scale) * W / W_norms

    def __call__(self, x, condition=None):
        "returns output y, and components of weight matrix needed log_det component (n_blocks, block_shape[0], block_shape[1])"
        W = self.get_normalised_weights()
        if condition is not None:
            x = jnp.concatenate((condition, x))
        y = W @ x + self.bias
        jac_3d = W[self.b_diag_mask_idxs].reshape(self.n_blocks, *self.block_shape)
        return y, jnp.log(jac_3d)

    @property
    def b_diag_mask(self):
        return jax.lax.stop_gradient(self._b_diag_mask)

    @property
    def b_diag_mask_idxs(self):
        return jax.lax.stop_gradient(self._b_diag_mask_idxs)

    @property
    def b_tril_mask(self):
        return jax.lax.stop_gradient(self._b_tril_mask)


def logmatmulexp(x, y):
    """
    Numerically stable version of ``(x.log() @ y.log()).exp()``. From numpyro https://github.com/pyro-ppl/numpyro/blob/f2ff89a3a7147617e185eb51148eb15d56d44661/numpyro/distributions/util.py#L387
    """
    x_shift = lax.stop_gradient(jnp.amax(x, -1, keepdims=True))
    y_shift = lax.stop_gradient(jnp.amax(y, -2, keepdims=True))
    xy = jnp.log(jnp.matmul(jnp.exp(x - x_shift), jnp.exp(y - y_shift)))
    return xy + x_shift + y_shift


class TanhBNAF:
    """
    Tanh transformation compatible with BNAF (log_abs_det provided as 3D array).
    Condition is ignored. Output shape is (n_blocks, *block_size), where
    output[i] is the log jacobian for the ii-th block.
    """

    def __init__(self, n_blocks: int):
        self.n_blocks = n_blocks

    def __call__(self, x):
        d = x.shape[0] // self.n_blocks
        log_det_vals = -2 * (x + jax.nn.softplus(-2 * x) - jnp.log(2.0))
        log_det = jnp.full((self.n_blocks, d, d), -jnp.inf)
        log_det = log_det.at[:, jnp.arange(d), jnp.arange(d)].set(
            log_det_vals.reshape(self.n_blocks, d)
        )
        return jnp.tanh(x), log_det


class BlockAutoregressiveNetwork(eqx.Module, Bijection):
    n_layers: int
    layers: list
    activation: Callable

    def __init__(
        self,
        key: random.PRNGKey,
        dim: int,
        condition_dim: int = 0,
        n_layers: int = 3,
        block_size: tuple = (8, 8),
        activation=TanhBNAF,
    ):

        self.n_layers = n_layers

        layers = []

        block_sizes = [
            (block_size[0], 1),
            *[block_size] * (n_layers - 2),
            (1, block_size[1]),
        ]
        condition_dims = [condition_dim if i==0 else 0 for i in range(n_layers)]
        for size, c_d in zip(block_sizes, condition_dims):
            key, subkey = random.split(key)
            layers.extend(
                [BlockAutoregressiveLinear(subkey, dim, size, c_d), activation(dim)]
            )
        self.layers = layers[:-1]
        self.activation = activation

    def transform(self, x: jnp.ndarray, condition=None):
        y = self.layers[0](x, condition)[0]
        for layer in self.layers[1:]:
            y = layer(y)[0]
        return y

    def transform_and_log_abs_det_jacobian(self, x: jnp.ndarray, condition=None):
        y, log_det_3d_0 = self.layers[0](x, condition)
        log_det_3ds = [log_det_3d_0]

        for layer in self.layers[1:]:
            y, log_det_3d = layer(y)
            log_det_3ds.append(log_det_3d)

        logdet = log_det_3ds[-1]
        for ld in reversed(log_det_3ds[:-1]):
            logdet = logmatmulexp(logdet, ld)
        return y, logdet.sum()

    def inverse(*args, **kwargs):
        return NotImplementedError(
            "This transform would require numerical methods for inversion."
        )



from flowjax.bijections.abc import ParameterisedBijection
import jax.numpy as jnp

class Affine(ParameterisedBijection):
    "Affine transformation compatible with neural network parameterisation."
    def transform(self, x, loc, log_scale):
        return x * jnp.exp(log_scale) + loc

    def transform_and_log_abs_det_jacobian(self, x, loc, log_scale):
        return x * jnp.exp(log_scale) + loc, jnp.sum(log_scale)

    def inverse(self, y, loc, log_scale):
        return (y - loc) / jnp.exp(log_scale)

    def num_params(self, dim: int):
        return dim * 2

    def get_args(self, params):
        loc, log_scale = params.split(2)
        return loc, log_scale



from abc import ABC, abstractmethod
import jax.numpy as jnp
from typing import Optional


class Bijection(ABC):
    """Basic bijection class. All bijections should support conditioning variables
    (even if ignored)."""

    @abstractmethod
    def transform(self, x: jnp.ndarray, condition: Optional[jnp.ndarray] = None):
        """Apply transformation."""
        pass

    @abstractmethod
    def transform_and_log_abs_det_jacobian(
        self, x: jnp.ndarray, condition: Optional[jnp.ndarray] = None
    ):
        """Apply transformation and compute log absolute value of the Jacobian determinant."""
        pass

    @abstractmethod
    def inverse(self, y: jnp.ndarray, condition: Optional[jnp.ndarray] = None):
        """Invert the transformation."""
        pass


class ParameterisedBijection(ABC):
    """Bijection which facilitates parameterisation with a
    neural network output (e.g. as in coupling flows)."""

    @abstractmethod
    def transform(self, x: jnp.ndarray, *args):
        """Apply transformation."""
        pass

    @abstractmethod
    def transform_and_log_abs_det_jacobian(self, x: jnp.ndarray, *args):
        """Apply transformation and compute log absolute value of the Jacobian determinant."""
        pass

    @abstractmethod
    def inverse(self, y: jnp.ndarray, *args):
        """Invert the transformation."""
        pass

    @abstractmethod
    def num_params(self, dim: int):
        "Total number of parameters required for bijection."
        pass

    @abstractmethod
    def get_args(self, params: jnp.ndarray):
        "Transform unconstrained vector of params (e.g. nn output) into args for transformation."
        pass



from flowjax.flows import Flow
from jax import random
import jax.numpy as jnp
import equinox as eqx
import optax
from tqdm import tqdm
from typing import Optional


def train_flow(
    key: random.PRNGKey,
    flow: Flow,
    x: jnp.ndarray,
    condition: Optional[jnp.ndarray] = None,
    max_epochs: int = 50,
    max_patience: int = 5,
    learning_rate: float = 5e-4,
    batch_size: int = 256,
    val_prop: float = 0.1,
    show_progress: bool = True,
):
    """Train flow with Adam optimizer.

    Args:
        key (random.PRNGKey): Jax key.
        flow (Flow): Flow to train.
        x (jnp.ndarray): Samples from the target distribution (each row being a sample).
        condition (Optional[jnp.ndarray], optional): Conditioning variables corresponding to x if learning a conditional distribution. Defaults to None.
        max_epochs (int, optional): Maximum number of epochs. Defaults to 50.
        max_patience (int, optional): Number of consecutive epochs with no validation loss improvement after which training is terminated. Defaults to 5.
        learning_rate (float, optional): Adam learning rate. Defaults to 5e-4.
        batch_size (int, optional): Batch size. Defaults to 256.
        val_prop (float, optional): Proportion of data to use for validation. Defaults to 0.1.
        show_progress (bool, optional): Whether to show progress bar. Defaults to True.
    """
    def loss(flow, x, condition=None):
        return -flow.log_prob(x, condition).mean()

    @eqx.filter_jit
    def step(flow, optimizer, opt_state, x, condition=None):
        loss_val, grads = eqx.filter_value_and_grad(loss)(flow, x, condition)
        updates, opt_state = optimizer.update(grads, opt_state)
        flow = eqx.apply_updates(flow, updates)
        return flow, opt_state, loss_val

    key, subkey = random.split(key)

    inputs = (x,) if condition is None else (x, condition)
    train_args, val_args = train_val_split(subkey, inputs, val_prop=val_prop)

    optimizer = optax.adam(learning_rate=learning_rate)

    best_params, static = eqx.partition(flow, eqx.is_array)

    opt_state = optimizer.init(best_params)

    losses = {"train": [], "val": []}

    loop = tqdm(range(max_epochs)) if show_progress is True else range(max_epochs)
    for epoch in loop:
        key, subkey = random.split(key)
        train_args = random_permutation_multiple(subkey, train_args)

        epoch_train_loss = 0
        batches = range(0, train_args[0].shape[0] - batch_size, batch_size)
        for i in batches:
            batch = tuple(a[i : i + batch_size] for a in train_args)
            flow, opt_state, loss_i = step(flow, optimizer, opt_state, *batch)
            epoch_train_loss += loss_i.item() / len(batches)

        epoch_val_loss = 0
        batches = range(0, val_args[0].shape[0] - batch_size, batch_size)
        for i in batches:
            batch = tuple(a[i : i + batch_size] for a in val_args)
            epoch_val_loss += loss(flow, *batch).item() / len(batches)

        losses["train"].append(epoch_train_loss)
        losses["val"].append(epoch_val_loss)

        if epoch_val_loss == min(losses["val"]):
            best_params, _ = eqx.partition(flow, eqx.is_array)

        elif count_fruitless(losses["val"]) > max_patience:
            print("Max patience reached.")
            break

        if show_progress:
            loop.set_postfix({k: v[-1] for k, v in losses.items()})

    flow = eqx.combine(best_params, static)
    return flow, losses


def train_val_split(key: random.PRNGKey, arrays, val_prop: float = 0.1):
    "Returns ((train_x, train_y), (val_x, val_y), ...)). Split on axis 0."
    assert 0 <= val_prop <= 1
    key, subkey = random.split(key)
    arrays = random_permutation_multiple(subkey, arrays)
    n_val = round(val_prop * arrays[0].shape[0])
    train = tuple(a[:-n_val] for a in arrays)
    val = tuple(a[-n_val:] for a in arrays)
    return train, val


def random_permutation_multiple(key, arrays):
    "Randomly permute multiple arrays on axis 0 (consistent between arrays)."
    n = arrays[0].shape[0]
    shuffle = random.permutation(key, jnp.arange(n))
    arrays = tuple(a[shuffle] for a in arrays)
    return arrays


def count_fruitless(losses: list):
    """Given a list of losses from each epoch, count the number of epochs since
    the minimum loss"""
    min_idx = jnp.array(losses).argmin().item()
    return len(losses) - min_idx - 1



from flowjax.bijections.abc import Bijection
import jax.numpy as jnp
import equinox as eqx
from jax import random
from typing import Optional, List


class Chain(Bijection, eqx.Module):
    bijections: List[Bijection]

    def __init__(self, bijections: List[Bijection]):
        """Chain together bijections to form another bijection.

        Args:
            bijections (List[Bijection]): List of bijections.
        """
        self.bijections = bijections

    def transform(self, x, condition=None):
        z = x
        for bijection in self.bijections:
            z = bijection.transform(z, condition)
        return z

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        log_abs_det_jac = 0
        z = x
        for bijection in self.bijections:
            z, log_abs_det_jac_i = bijection.transform_and_log_abs_det_jacobian(
                z, condition
            )
            log_abs_det_jac += log_abs_det_jac_i
        return z, log_abs_det_jac

    def inverse(self, z: jnp.ndarray, condition=None):
        x = z
        for bijection in reversed(self.bijections):
            x = bijection.inverse(x, condition)
        return x


class Permute(Bijection):
    permutation: jnp.ndarray
    inverse_permutation: jnp.ndarray

    def __init__(self, permutation: jnp.ndarray):
        """Permutation transformation. Note condition is ignored.

        Args:
            permutation (jnp.ndarray): Indexes 0-(dim-1) representing new order.
        """
        assert (permutation.sort() == jnp.arange(len(permutation))).all()
        self.permutation = permutation
        self.inverse_permutation = jnp.argsort(permutation)

    def transform(self, x, condition=None):
        return x[self.permutation]

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        return x[self.permutation], jnp.array(0)

    def inverse(self, y, condition=None):
        return y[self.inverse_permutation]


class Flip(Bijection):
    """Flip the input array. Condition argument is ignored."""

    def transform(self, x, condition=None):
        return jnp.flip(x)

    def transform_and_log_abs_det_jacobian(self, x, condition=None):
        return jnp.flip(x), jnp.array(0)

    def inverse(self, y, condition=None):
        return jnp.flip(y)


def intertwine_permute(
    bijection_list: List[Bijection],
    strategy: str,
    key: Optional[random.PRNGKey] = None,
    dim: Optional[int] = None,
):
    """Given a list of bijections, add permutations between layers. i.e.
    with bijections [a,b,c], returns [a, perm1, b, perm2, c].

    Args:
        bijection_list (list[Bijection]): List of bijections.
        strategy (str): Either "flip" or "random".
        key (random.PRNGKey, optional): Random key (ignored for flip). Defaults to None.
        dim (int, optional): Dimension (ignored for flip). Defaults to None.
    """
    bijections = []
    if strategy.lower() == "flip":
        for b in bijection_list:
            bijections.extend([b, Flip()])
        return bijections[:-1]

    elif strategy.lower() == "random":
        permutations = jnp.row_stack(
            [jnp.arange(dim) for _ in range(len(bijection_list))]
        )
        permutations = random.permutation(key, permutations, 1, True)
        bijections = []

        for bijection, permutation in zip(bijection_list, permutations):
            bijections.extend([bijection, Permute(permutation)])
        return bijections[:-1]

    else:
        ValueError("permute strategy should be 'flip' or 'random'")
